=== modified file 'NEWS'
--- NEWS	2007-06-04 18:59:21 +0000
+++ NEWS	2007-06-06 01:07:25 +0000
@@ -29,6 +29,10 @@
       understanding what changes have occurred
       (John Arbash Meinel, #83887)
 
+    * TreeTransform avoids many renames when contructing large trees,
+      improving speed.  3.25x speedups have been observed for construction of
+      kernel-sized-trees, and checkouts are 1.28x faster.  (Aaron Bentley)
+
   BUGFIXES:
 
     * ``bzr push`` should only connect to the remote location one time.
@@ -50,6 +54,9 @@
     * WorkingTree.remove works correctly with tree references, and when pwd is
       not the tree root. (Aaron Bentley)
 
+    * Merge no longer fails when a file is renamed in one tree and deleted
+      in the other. (Aaron Bentley, #110279)
+
 bzr 0.16  2007-05-07
   
   BUGFIXES:

=== modified file 'bzrlib/merge.py'
--- bzrlib/merge.py	2007-03-21 01:34:41 +0000
+++ bzrlib/merge.py	2007-06-06 01:07:25 +0000
@@ -870,7 +870,7 @@
             except KeyError:
                 this_name = other_name = self.tt.final_name(trans_id)
             other_path = fp.get_path(trans_id)
-            if this_parent is not None:
+            if this_parent is not None and this_name is not None:
                 this_parent_path = \
                     fp.get_path(self.tt.trans_id_file_id(this_parent))
                 this_path = pathjoin(this_parent_path, this_name)

=== modified file 'bzrlib/tests/test_merge.py'
--- bzrlib/tests/test_merge.py	2007-03-06 10:22:33 +0000
+++ bzrlib/tests/test_merge.py	2007-06-06 01:50:55 +0000
@@ -249,3 +249,24 @@
         registry.remove('merge4')
         self.assertFalse('merge4' in [x[0] for x in 
                         merge_type_option.iter_switches()])
+
+    def test_merge_other_moves_we_deleted(self):
+        tree_a = self.make_branch_and_tree('A')
+        tree_a.lock_write()
+        self.addCleanup(tree_a.unlock)
+        self.build_tree(['A/a'])
+        tree_a.add('a')
+        tree_a.commit('1', rev_id='rev-1')
+        tree_a.flush()
+        tree_a.rename_one('a', 'b')
+        tree_a.commit('2')
+        bzrdir_b = tree_a.bzrdir.sprout('B', revision_id='rev-1')
+        tree_b = bzrdir_b.open_workingtree()
+        tree_b.lock_write()
+        self.addCleanup(tree_b.unlock)
+        os.unlink('B/a')
+        tree_b.commit('3')
+        try:
+            tree_b.merge_from_branch(tree_a.branch)
+        except AttributeError:
+            self.fail('tried to join a path when name was None')

=== modified file 'bzrlib/tests/test_transform.py'
--- bzrlib/tests/test_transform.py	2007-03-11 20:34:48 +0000
+++ bzrlib/tests/test_transform.py	2007-06-06 01:28:18 +0000
@@ -394,7 +394,6 @@
         self.assertEqual(os.readlink(self.wt.abspath('oz/wizard')),
                          'wizard-target')
 
-
     def get_conflicted(self):
         create,root = self.get_transform()
         create.new_file('dorothy', root, 'dorothy', 'dorothy-id')
@@ -545,9 +544,11 @@
         wt = transform._tree
         transform.new_file('set_on_creation', root, 'Set on creation', 'soc',
                            True)
-        sac = transform.new_file('set_after_creation', root, 'Set after creation', 'sac')
+        sac = transform.new_file('set_after_creation', root,
+                                 'Set after creation', 'sac')
         transform.set_executability(True, sac)
-        uws = transform.new_file('unset_without_set', root, 'Unset badly', 'uws')
+        uws = transform.new_file('unset_without_set', root, 'Unset badly',
+                                 'uws')
         self.assertRaises(KeyError, transform.set_executability, None, uws)
         transform.apply()
         self.assertTrue(wt.is_executable('soc'))
@@ -774,6 +775,174 @@
         finally:
             transform.finalize()
 
+    def test_rename_count(self):
+        transform, root = self.get_transform()
+        transform.new_file('name1', root, 'contents')
+        self.assertEqual(transform.rename_count, 0)
+        transform.apply()
+        self.assertEqual(transform.rename_count, 1)
+        transform2, root = self.get_transform()
+        transform2.adjust_path('name2', root,
+                               transform2.trans_id_tree_path('name1'))
+        self.assertEqual(transform2.rename_count, 0)
+        transform2.apply()
+        self.assertEqual(transform2.rename_count, 2)
+
+    def test_change_parent(self):
+        """Ensure that after we change a parent, the results are still right.
+
+        Renames and parent changes on pending transforms can happen as part
+        of conflict resolution, and are explicitly permitted by the
+        TreeTransform API.
+
+        This test ensures they work correctly with the rename-avoidance
+        optimization.
+        """
+        transform, root = self.get_transform()
+        parent1 = transform.new_directory('parent1', root)
+        child1 = transform.new_file('child1', parent1, 'contents')
+        parent2 = transform.new_directory('parent2', root)
+        transform.adjust_path('child1', parent2, child1)
+        transform.apply()
+        self.failIfExists(self.wt.abspath('parent1/child1'))
+        self.failUnlessExists(self.wt.abspath('parent2/child1'))
+        # rename limbo/new-1 => parent1, rename limbo/new-3 => parent2
+        # no rename for child1 (counting only renames during apply)
+        self.failUnlessEqual(2, transform.rename_count)
+
+    def test_cancel_parent(self):
+        """Cancelling a parent doesn't cause deletion of a non-empty directory
+
+        This is like the test_change_parent, except that we cancel the parent
+        before adjusting the path.  The transform must detect that the
+        directory is non-empty, and move children to safe locations.
+        """
+        transform, root = self.get_transform()
+        parent1 = transform.new_directory('parent1', root)
+        child1 = transform.new_file('child1', parent1, 'contents')
+        child2 = transform.new_file('child2', parent1, 'contents')
+        try:
+            transform.cancel_creation(parent1)
+        except OSError:
+            self.fail('Failed to move child1 before deleting parent1')
+        transform.cancel_creation(child2)
+        transform.create_directory(parent1)
+        try:
+            transform.cancel_creation(parent1)
+        # If the transform incorrectly believes that child2 is still in
+        # parent1's limbo directory, it will try to rename it and fail
+        # because was already moved by the first cancel_creation.
+        except OSError:
+            self.fail('Transform still thinks child2 is a child of parent1')
+        parent2 = transform.new_directory('parent2', root)
+        transform.adjust_path('child1', parent2, child1)
+        transform.apply()
+        self.failIfExists(self.wt.abspath('parent1'))
+        self.failUnlessExists(self.wt.abspath('parent2/child1'))
+        # rename limbo/new-3 => parent2, rename limbo/new-2 => child1
+        self.failUnlessEqual(2, transform.rename_count)
+
+    def test_adjust_and_cancel(self):
+        """Make sure adjust_path keeps track of limbo children properly"""
+        transform, root = self.get_transform()
+        parent1 = transform.new_directory('parent1', root)
+        child1 = transform.new_file('child1', parent1, 'contents')
+        parent2 = transform.new_directory('parent2', root)
+        transform.adjust_path('child1', parent2, child1)
+        transform.cancel_creation(child1)
+        try:
+            transform.cancel_creation(parent1)
+        # if the transform thinks child1 is still in parent1's limbo
+        # directory, it will attempt to move it and fail.
+        except OSError:
+            self.fail('Transform still thinks child1 is a child of parent1')
+        transform.finalize()
+
+    def test_noname_contents(self):
+        """TreeTransform should permit deferring naming files."""
+        transform, root = self.get_transform()
+        parent = transform.trans_id_file_id('parent-id')
+        try:
+            transform.create_directory(parent)
+        except KeyError:
+            self.fail("Can't handle contents with no name")
+        transform.finalize()
+
+    def test_noname_contents_nested(self):
+        """TreeTransform should permit deferring naming files."""
+        transform, root = self.get_transform()
+        parent = transform.trans_id_file_id('parent-id')
+        try:
+            transform.create_directory(parent)
+        except KeyError:
+            self.fail("Can't handle contents with no name")
+        child = transform.new_directory('child', parent)
+        transform.adjust_path('parent', root, parent)
+        transform.apply()
+        self.failUnlessExists(self.wt.abspath('parent/child'))
+        self.assertEqual(1, transform.rename_count)
+
+    def test_reuse_name(self):
+        """Avoid reusing the same limbo name for different files"""
+        transform, root = self.get_transform()
+        parent = transform.new_directory('parent', root)
+        child1 = transform.new_directory('child', parent)
+        try:
+            child2 = transform.new_directory('child', parent)
+        except OSError:
+            self.fail('Tranform tried to use the same limbo name twice')
+        transform.adjust_path('child2', parent, child2)
+        transform.apply()
+        # limbo/new-1 => parent, limbo/new-3 => parent/child2
+        # child2 is put into top-level limbo because child1 has already
+        # claimed the direct limbo path when child2 is created.  There is no
+        # advantage in renaming files once they're in top-level limbo, except
+        # as part of apply.
+        self.assertEqual(2, transform.rename_count)
+
+    def test_reuse_when_first_moved(self):
+        """Don't avoid direct paths when it is safe to use them"""
+        transform, root = self.get_transform()
+        parent = transform.new_directory('parent', root)
+        child1 = transform.new_directory('child', parent)
+        transform.adjust_path('child1', parent, child1)
+        child2 = transform.new_directory('child', parent)
+        transform.apply()
+        # limbo/new-1 => parent
+        self.assertEqual(1, transform.rename_count)
+
+    def test_reuse_after_cancel(self):
+        """Don't avoid direct paths when it is safe to use them"""
+        transform, root = self.get_transform()
+        parent2 = transform.new_directory('parent2', root)
+        child1 = transform.new_directory('child1', parent2)
+        transform.cancel_creation(parent2)
+        transform.create_directory(parent2)
+        child2 = transform.new_directory('child1', parent2)
+        transform.adjust_path('child2', parent2, child1)
+        transform.apply()
+        # limbo/new-1 => parent2, limbo/new-2 => parent2/child1
+        self.assertEqual(2, transform.rename_count)
+
+    def test_finalize_order(self):
+        """Finalize must be done in child-to-parent order"""
+        transform, root = self.get_transform()
+        parent = transform.new_directory('parent', root)
+        child = transform.new_directory('child', parent)
+        try:
+            transform.finalize()
+        except OSError:
+            self.fail('Tried to remove parent before child1')
+
+    def test_cancel_with_cancelled_child_should_succeed(self):
+        transform, root = self.get_transform()
+        parent = transform.new_directory('parent', root)
+        child = transform.new_directory('child', parent)
+        transform.cancel_creation(child)
+        transform.cancel_creation(parent)
+        transform.finalize()
+
+
 class TransformGroup(object):
     def __init__(self, dirname, root_id):
         self.name = dirname
@@ -1115,6 +1284,23 @@
         self.assertRaises(errors.WorkingTreeAlreadyPopulated, 
             build_tree, source.basis_tree(), target)
 
+    def test_build_tree_rename_count(self):
+        source = self.make_branch_and_tree('source')
+        self.build_tree(['source/file1', 'source/dir1/'])
+        source.add(['file1', 'dir1'])
+        source.commit('add1')
+        target1 = self.make_branch_and_tree('target1')
+        transform_result = build_tree(source.basis_tree(), target1)
+        self.assertEqual(2, transform_result.rename_count)
+
+        self.build_tree(['source/dir1/file2'])
+        source.add(['dir1/file2'])
+        source.commit('add3')
+        target2 = self.make_branch_and_tree('target2')
+        transform_result = build_tree(source.basis_tree(), target2)
+        # children of non-root directories should not be renamed
+        self.assertEqual(2, transform_result.rename_count)
+
 
 class MockTransform(object):
 
@@ -1127,6 +1313,7 @@
                 return True
         return False
 
+
 class MockEntry(object):
     def __init__(self):
         object.__init__(self)

=== modified file 'bzrlib/transform.py'
--- bzrlib/transform.py	2007-06-01 05:43:11 +0000
+++ bzrlib/transform.py	2007-06-06 04:10:17 +0000
@@ -51,16 +51,21 @@
 
 
 class _TransformResults(object):
-    def __init__(self, modified_paths):
+    def __init__(self, modified_paths, rename_count):
         object.__init__(self)
         self.modified_paths = modified_paths
+        self.rename_count = rename_count
 
 
 class TreeTransform(object):
     """Represent a tree transformation.
     
     This object is designed to support incremental generation of the transform,
-    in any order.  
+    in any order.
+
+    However, it gives optimum performance when parent directories are created
+    before their contents.  The transform is then able to put child files
+    directly in their parent directory, avoiding later renames.
     
     It is easy to produce malformed transforms, but they are generally
     harmless.  Attempting to apply a malformed transform will cause an
@@ -84,7 +89,8 @@
     def __init__(self, tree, pb=DummyProgress()):
         """Note: a tree_write lock is taken on the tree.
         
-        Use TreeTransform.finalize() to release the lock
+        Use TreeTransform.finalize() to release the lock (can be omitted if
+        TreeTransform.apply() called).
         """
         object.__init__(self)
         self._tree = tree
@@ -106,6 +112,15 @@
         self._new_name = {}
         self._new_parent = {}
         self._new_contents = {}
+        # A mapping of transform ids to their limbo filename
+        self._limbo_files = {}
+        # A mapping of transform ids to a set of the transform ids of children
+        # that their limbo directory has
+        self._limbo_children = {}
+        # Map transform ids to maps of child filename to child transform id
+        self._limbo_children_names = {}
+        # List of transform ids that need to be renamed from limbo into place
+        self._needs_rename = set()
         self._removed_contents = set()
         self._new_executability = {}
         self._new_reference_revision = {}
@@ -115,13 +130,14 @@
         self._removed_id = set()
         self._tree_path_ids = {}
         self._tree_id_paths = {}
+        # Cache of realpath results, to speed up canonical_path
         self._realpaths = {}
-        # Cache of realpath results, to speed up canonical_path
+        # Cache of relpath results, to speed up canonical_path
         self._relpaths = {}
-        # Cache of relpath results, to speed up canonical_path
         self._new_root = self.trans_id_tree_file_id(tree.get_root_id())
         self.__done = False
         self._pb = pb
+        self.rename_count = 0
 
     def __get_root(self):
         return self._new_root
@@ -129,12 +145,18 @@
     root = property(__get_root)
 
     def finalize(self):
-        """Release the working tree lock, if held, clean up limbo dir."""
+        """Release the working tree lock, if held, clean up limbo dir.
+
+        This is required if apply has not been invoked, but can be invoked
+        even after apply.
+        """
         if self._tree is None:
             return
         try:
-            for trans_id, kind in self._new_contents.iteritems():
-                path = self._limbo_name(trans_id)
+            entries = [(self._limbo_name(t), t, k) for t, k in
+                       self._new_contents.iteritems()]
+            entries.sort(reverse=True)
+            for path, trans_id, kind in entries:
                 if kind == "directory":
                     os.rmdir(path)
                 else:
@@ -165,8 +187,33 @@
         """Change the path that is assigned to a transaction id."""
         if trans_id == self._new_root:
             raise CantMoveRoot
+        previous_parent = self._new_parent.get(trans_id)
+        previous_name = self._new_name.get(trans_id)
         self._new_name[trans_id] = name
         self._new_parent[trans_id] = parent
+        if (trans_id in self._limbo_files and
+            trans_id not in self._needs_rename):
+            self._rename_in_limbo([trans_id])
+            self._limbo_children[previous_parent].remove(trans_id)
+            del self._limbo_children_names[previous_parent][previous_name]
+
+    def _rename_in_limbo(self, trans_ids):
+        """Fix limbo names so that the right final path is produced.
+
+        This means we outsmarted ourselves-- we tried to avoid renaming
+        these files later by creating them with their final names in their
+        final parents.  But now the previous name or parent is no longer
+        suitable, so we have to rename them.
+
+        Even for trans_ids that have no new contents, we must remove their
+        entries from _limbo_files, because they are now stale.
+        """
+        for trans_id in trans_ids:
+            old_path = self._limbo_files.pop(trans_id)
+            if trans_id not in self._new_contents:
+                continue
+            new_path = self._limbo_name(trans_id)
+            os.rename(old_path, new_path)
 
     def adjust_root_path(self, name, parent):
         """Emulate moving the root by moving all children, instead.
@@ -330,6 +377,13 @@
     def cancel_creation(self, trans_id):
         """Cancel the creation of new file contents."""
         del self._new_contents[trans_id]
+        children = self._limbo_children.get(trans_id)
+        # if this is a limbo directory with children, move them before removing
+        # the directory
+        if children is not None:
+            self._rename_in_limbo(children)
+            del self._limbo_children[trans_id]
+            del self._limbo_children_names[trans_id]
         delete_any(self._limbo_name(trans_id))
 
     def delete_contents(self, trans_id):
@@ -734,6 +788,8 @@
         
         If filesystem or inventory conflicts are present, MalformedTransform
         will be thrown.
+
+        If apply succeeds, finalize is not necessary.
         """
         conflicts = self.find_conflicts()
         if len(conflicts) != 0:
@@ -751,11 +807,40 @@
         self._tree.apply_inventory_delta(inventory_delta)
         self.__done = True
         self.finalize()
-        return _TransformResults(modified_paths)
+        return _TransformResults(modified_paths, self.rename_count)
 
     def _limbo_name(self, trans_id):
         """Generate the limbo name of a file"""
-        return pathjoin(self._limbodir, trans_id)
+        limbo_name = self._limbo_files.get(trans_id)
+        if limbo_name is not None:
+            return limbo_name
+        parent = self._new_parent.get(trans_id)
+        # if the parent directory is already in limbo (e.g. when building a
+        # tree), choose a limbo name inside the parent, to reduce further
+        # renames.
+        use_direct_path = False
+        if self._new_contents.get(parent) == 'directory':
+            filename = self._new_name.get(trans_id)
+            if filename is not None:
+                if parent not in self._limbo_children:
+                    self._limbo_children[parent] = set()
+                    self._limbo_children_names[parent] = {}
+                    use_direct_path = True
+                # the direct path can only be used if no other file has
+                # already taken this pathname, i.e. if the name is unused, or
+                # if it is already associated with this trans_id.
+                elif (self._limbo_children_names[parent].get(filename)
+                      in (trans_id, None)):
+                    use_direct_path = True
+        if use_direct_path:
+            limbo_name = pathjoin(self._limbo_files[parent], filename)
+            self._limbo_children[parent].add(trans_id)
+            self._limbo_children_names[parent][filename] = trans_id
+        else:
+            limbo_name = pathjoin(self._limbodir, trans_id)
+            self._needs_rename.add(trans_id)
+        self._limbo_files[trans_id] = limbo_name
+        return limbo_name
 
     def _apply_removals(self, inv, inventory_delta):
         """Perform tree operations that remove directory/inventory names.
@@ -781,6 +866,8 @@
                     except OSError, e:
                         if e.errno != errno.ENOENT:
                             raise
+                    else:
+                        self.rename_count += 1
                 if trans_id in self._removed_id:
                     if trans_id == self._new_root:
                         file_id = self._tree.inventory.root.file_id
@@ -812,12 +899,15 @@
                 if trans_id in self._new_contents or \
                     self.path_changed(trans_id):
                     full_path = self._tree.abspath(path)
-                    try:
-                        os.rename(self._limbo_name(trans_id), full_path)
-                    except OSError, e:
-                        # We may be renaming a dangling inventory id
-                        if e.errno != errno.ENOENT:
-                            raise
+                    if trans_id in self._needs_rename:
+                        try:
+                            os.rename(self._limbo_name(trans_id), full_path)
+                        except OSError, e:
+                            # We may be renaming a dangling inventory id
+                            if e.errno != errno.ENOENT:
+                                raise
+                        else:
+                            self.rename_count += 1
                     if trans_id in self._new_contents:
                         modified_paths.append(full_path)
                         del self._new_contents[trans_id]
@@ -1151,9 +1241,9 @@
     top_pb = bzrlib.ui.ui_factory.nested_progress_bar()
     pp = ProgressPhase("Build phase", 2, top_pb)
     if tree.inventory.root is not None:
-        # this is kindof a hack: we should be altering the root 
-        # as partof the regular tree shape diff logic.
-        # the conditional test hereis to avoid doing an
+        # This is kind of a hack: we should be altering the root
+        # as part of the regular tree shape diff logic.
+        # The conditional test here is to avoid doing an
         # expensive operation (flush) every time the root id
         # is set within the tree, nor setting the root and thus
         # marking the tree as dirty, because we use two different
@@ -1219,10 +1309,11 @@
             wt.add_conflicts(conflicts)
         except errors.UnsupportedOperation:
             pass
-        tt.apply()
+        result = tt.apply()
     finally:
         tt.finalize()
         top_pb.finished()
+    return result
 
 
 def _reparent_children(tt, old_parent, new_parent):
@@ -1469,9 +1560,9 @@
                 if kind[0] == 'file' and (backups or kind[1] is None):
                     wt_sha1 = working_tree.get_file_sha1(file_id)
                     if merge_modified.get(file_id) != wt_sha1:
-                        # acquire the basis tree lazyily to prevent the expense
-                        # of accessing it when its not needed ? (Guessing, RBC,
-                        # 200702)
+                        # acquire the basis tree lazily to prevent the
+                        # expense of accessing it when it's not needed ?
+                        # (Guessing, RBC, 200702)
                         if basis_tree is None:
                             basis_tree = working_tree.basis_tree()
                             basis_tree.lock_read()

=== modified file 'doc/developers/performance-roadmap.txt'
--- doc/developers/performance-roadmap.txt	2007-06-06 07:11:52 +0000
+++ doc/developers/performance-roadmap.txt	2007-06-06 07:45:14 +0000
@@ -16,6 +16,8 @@
 
 .. include:: performance-roadmap-rationale.txt
 
+.. include:: planned-performance-changes.txt
+
 Analysis of use cases
 #####################
 

=== added file 'doc/developers/planned-performance-changes.txt'
--- doc/developers/planned-performance-changes.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/planned-performance-changes.txt	2007-06-06 07:45:14 +0000
@@ -0,0 +1,174 @@
+Planned changes to the bzr core
+===============================
+
+Delivering the best possible performance requires changing the bzr core design
+from that present in 0.16. Some of these changes are incremental and can be
+done with no impact on disk format. Many of them however do require changes to
+the disk format, and these can be broken into two sets of changes, those which
+are sufficiently close to the model bzr uses today to interoperate with the
+0.16 disk formats, and those that are not able to interoperate with the 0.16
+disk formats - specifically some planned changes may result in data which
+cannot be exported to bzr 0.16's disk formats and then imported back to the new
+format without losing critical information. If/when this takes place it will be
+essentially a migration for users to switch from their bzr 0.16 repository to a
+bzr that supports them. We plan to batch all such changes into one large
+'experimental' repository format, which will be complete stable and usable
+before we migrate it to become a supported format. Getting new versions of bzr
+in widespread use at that time will be very important, otherwise the user base
+may be split in two - users that have upgraded and users that have not.
+
+The following changes are grouped according to their compatability impact:
+library only, disk format but interoperable, disk format interoperability
+unknown, and disk format, not interoperable.
+
+Library changes
+---------------
+
+These changes will change bzrlib's API but will not affect the disk format and
+thus do not pose a significant migration issue.
+
+ * For our 20 core use cases, we plan to add targeted API's to bzrlib that are
+   repository-representation agnostic. These will instead reflect the shape of
+   data access most optimal for that case.
+
+ * Deprecate 'versioned files' as a library concept. Instead of asking for
+   information about a file-over-time as a special case, we will move to an API
+   that assumes less coupling between the historical information and the
+   ability to obtain texts/deltas etc. Specifically, we need to remove all
+   API's that act in terms of on disk representation except those within a
+   given repository implementation.
+
+ * Create a validator for revisions that is more amenable to use by other parts
+   of the code base than just the gpg signing facility. This can be done today
+   without changing disk, possibly with a performance hit until the disk
+   formats match the validatory logic. It will be hard to tell if we have the
+   right routine for that until all the disk changes are complete, so while
+   this is a library only change, its likely one that will be delayed to near
+   the end of the process.
+
+ * Add an explicit API for managing cached annotations. While annotations are
+   considered a cache this is not exposed in such a way that cache operations
+   like 'drop the cache' can be performed. On current disk formats the cache is
+   mandatory, but an API to manage would allow refreshing of the cache (e.g.
+   after ghosts are filled in in baz conversions).
+
+ * Use the _iter_changes API to perform merges. This is a small change that may
+   remove the need to use inventories in merge, making a dramatic difference to
+   merge performance once the tree shape comparison optimisations are
+   implemented.
+
+ * Create a network-efficient revision graph API. This is the logic at the
+   start of push and pull operations, which currently scales O(graph size).
+   Fixing the scaling can be done, but there are tradeoffs to latency and
+   performance to consider, making it a little tricky to get right.
+
+ * Working tree disk operation ordering. We plan to change the order in which
+   some operations are done (specifically TreeTransform ones) to improve
+   performance. There is already a 66% performance boost in that area going
+   through review.
+
+ * Stop requiring full memory copies of files. Currently bzr requires that it
+   can hold 3 copies of any file its versioning in memory. Solving this is
+   tricky, particularly without performance regressions on small files, but
+   without solving it versioning of .iso and other large objects will continue
+   to be extremely painful.
+
+ * Add an API for per-file graph access that alllows incremental access and is
+   suitable for on-demand generation if desired.
+
+ * Repository stacking API. Allowing multiple databases to be stacked to give a
+   single 'repository' will allow implementation of some long desired features
+   like history horizons, and bundle usage where the bundle is not added to the
+   local repository just to examine its contents.
+
+ * Revision data manipulation API. We need a single streaming API for adding
+   data to or getting it from a repository. This will need to allow hints such
+   as 'optimise for size', or 'optimise for fast-addition' to meet the various
+   users planned, but it is a core part of the library today, and its not
+   sufficiently clean to let us simplify/remove a lot of related code today.
+
+Interoperable disk changes
+--------------------------
+
+ * New container format to allow single-file description of multiple named
+   objects. This will provide the basis for transmission of revisions over the
+   network, the new bundle format, and possibly a new repository format as
+   well.
+
+ * Separate the annotation cache from the storage of actual file texts and make
+   the annotation style, and when to do it, configurable. This will reduce data
+   sent over the wire when repositories have had 'needs-annotations' turned
+   off, which very large trees may choose to do - generating just-in-time
+   annotations may be desirable for those trees (even when performing
+   annotation based merges).
+
+ * Repository disk operation ordering. The order that tasks access data within
+   the repository and the layout of the data should be harmonised. This will
+   require disk format changes but does not inherently alter the model, so its
+   straight forward to export from a repository that has been optimised in this
+   way to a 0.16 based repository.
+
+ * Inventory representation. An inventory is a logical description of the shape
+   of a version controlled tree. Currently we operate on the whole inventory as
+   a tree broken down per directory, but we store it as a flat file. This scale
+   very poorly as even a minor change between inventories requires us to scan
+   the entire file, and in large trees this is many megabytes of data to
+   consider. We are investigating the exact form, but the intent is to change
+   the serialisation of inventories so that comparing two inventories can be
+   done in some smaller time - e.g. O(log N) scaling. Whatever form this takes,
+   a repository that can export it directly will be able to perform operations
+   between two historical trees much more efficiently than the current
+   repositories.
+
+ * Delta storage optimisation. We plan to change the delta storage logic to use
+   a binary delta like xdelta rather than using line based deltas from python.
+   These binary deltas could be done along ancestry ordering, or other
+   arbitrary patterns chosen for their intended use. Line based deltas will
+   still be created for cached annotations. This is still under some discussion.
+   http://bazaar-vcs.org/PerformanceRoadmap/Xdelta
+
+ * Greatest distance from origin cache. This is a possible change to introduce,
+   but it may be unnecessary - listed here for completeness till it has been
+   established as [un]needed.
+
+Possibly non-interoperable disk changes
+---------------------------------------
+
+ * Removing of derivable data from the core of bzr. Much of the data that bzr
+   stores is derivable from the users source files. For instance the
+   annotations that record who introduced a line. Given the full history for a
+   repository we can recreate that at any time. We want to remove the
+   dependence of the core of bzr on any data that is derivable, because doing
+   this will give us the freedom to:
+
+   * Improve the derivation algorithm over time.
+   * Deal with bugs in the derivation algorithms without having 'corrupt
+     repositories' or such things.
+
+   However, some of the data that is technically derived, like the per-file
+   merge graph, is both considered core, and can be generated differently when
+   certain circumstances arive, by bzr 0.16. Any change to the 'core' status of
+   that data will discard data that cannot be recreated and thus lead to the
+   inability to export from a format where that is derived data to bzr 0.16's
+   formats without errors occuring in those circumstances. Some of the data
+   that may be considered for this includes:
+
+   * Per file merge graphs
+   * Annotations
+
+Non-interoperable disk changes
+------------------------------
+
+ * Drop the per-file merge graph 'cache' currently held in the FILE-ID.kndx
+   files. A specific case of removing derivable data, this may allow smaller
+   inventory metadata and also make it easier to allow two different trees (in
+   terms of last-change made, e.g. if one is a working tree) to be compared
+   using a hash-tree style approach.
+
+ * Use hash based names for some objects in the bzr database. Because it would force
+   total-knowledge-of-history on the graph revision objects will not be namable
+   via hash's and neither will revisio signatures. Other than that though we
+   can in principle use hash's e.g. SHA1 for everything else. There are many
+   unanswered questions about hash based naming related to locality of
+   reference impacts, which need to be answered before this becomes a definite
+   item.

